// Generated by CoffeeScript 1.8.0
define(function() {
  var OfflineDb;
  return OfflineDb = (function() {
    function OfflineDb() {
      var request;
      request = (typeof indexedDB !== "undefined" && indexedDB !== null ? indexedDB : shimIndexedDB).open('dbms', 1);
      request.onsuccess = (function(_this) {
        return function(e) {
          var cb, _i, _len, _ref;
          _this.db = e.target.result;
          if (_this.readyCbs) {
            _ref = _this.readyCbs;
            for (_i = 0, _len = _ref.length; _i < _len; _i++) {
              cb = _ref[_i];
              cb();
            }
            return delete _this.readyCbs;
          }
        };
      })(this);
      request.onupgradeneeded = function(e) {
        var db;
        db = e.target.result;
        return db.createObjectStore('instances', {
          keyPath: '_key'
        });
      };
    }

    OfflineDb.prototype.available = function() {
      return localStorage.getItem('dbName');
    };

    OfflineDb.prototype.create = function(version, clientId, dbName, schema) {
      localStorage.setItem('version', version);
      localStorage.setItem('clientId', clientId);
      localStorage.setItem('dbName', dbName);
      return localStorage.setItem("" + dbName + ".schema", JSON.stringify(schema));
    };

    OfflineDb.prototype.migrate = function(fromVersion, toVersion) {};

    OfflineDb.prototype.storeChanges = function(db, changes) {
      return localStorage.setItem("" + db + ".changes", JSON.stringify(changes));
    };

    OfflineDb.prototype.storePendingUpdate = function(pendingUpdate) {
      return localStorage.setItem('pendingUpdate', JSON.stringify(pendingUpdate));
    };

    OfflineDb.prototype.clearPendingUpdate = function() {
      return localStorage.removeItem('pendingUpdate');
    };

    OfflineDb.prototype.clearChanges = function(db) {
      return localStorage.removeItem("" + db + ".changes");
    };

    OfflineDb.prototype.clear = function(session) {
      var db;
      localStorage.removeItem('version');
      localStorage.removeItem('clientId');
      this.clearPendingUpdate();
      db = localStorage.getItem('dbName');
      localStorage.removeItem('dbName');
      localStorage.removeItem("" + db + ".schema");
      localStorage.removeItem("" + db + ".changes");
      return (typeof indexedDB !== "undefined" && indexedDB !== null ? indexedDB : shimIndexedDB).deleteDatabase('dbms');
    };

    OfflineDb.prototype.write = function(data, cb) {
      return this.ready((function(_this) {
        return function() {
          var allModelData, changeKey, id, instanceChanges, instancesStore, modelData, modelName, resourceChanges, subChanges, _fn;
          allModelData = {};
          resourceChanges = {};
          for (changeKey in data) {
            subChanges = data[changeKey];
            if (changeKey[0] === '/') {
              resourceChanges[changeKey] = subChanges;
            } else {
              allModelData[changeKey] = subChanges;
            }
          }
          instancesStore = _this.db.transaction(['instances'], 'readwrite').objectStore('instances');
          for (modelName in allModelData) {
            modelData = allModelData[modelName];
            _fn = function(modelName, id, instanceChanges) {
              var key;
              key = "" + modelName + "." + id;
              if (instanceChanges === 'delete') {
                return instancesStore["delete"](key);
              } else {
                return instancesStore.get(key).onsuccess = function(e) {
                  var instanceData, _ref;
                  instanceData = (_ref = e.target.result) != null ? _ref : {};
                  if (!instanceData._key) {
                    instanceData._key = key;
                    instanceData._model = modelName;
                    instanceData._id = id;
                  }
                  _.extend(instanceData, instanceChanges);
                  return instancesStore.put(instanceData);
                };
              }
            };
            for (id in modelData) {
              instanceChanges = modelData[id];
              _fn(modelName, id, instanceChanges);
            }
          }
          return typeof cb === "function" ? cb() : void 0;
        };
      })(this));
    };

    OfflineDb.prototype.ready = function(cb) {
      if (this.db) {
        return cb();
      } else {
        if (this.readyCbs == null) {
          this.readyCbs = [];
        }
        return this.readyCbs.push(cb);
      }
    };

    OfflineDb.prototype.changeId = function(model, id, newId) {
      return this.ready((function(_this) {
        return function() {
          var instancesStore;
          instancesStore = _this.db.transaction(['instances'], 'readwrite').objectStore('instances');
          return instancesStore.get("" + model + "." + id).onsuccess = function(e) {
            var instanceData;
            instanceData = e.target.result;
            instancesStore["delete"]("" + model + "." + id);
            instanceData._id = newId;
            instanceData._key = "" + model + "." + newId;
            return instancesStore.put(instanceData);
          };
        };
      })(this));
    };

    OfflineDb.prototype.read = function(version, db, cb) {
      return this.ready((function(_this) {
        return function() {
          var cursor, data, instancesStore;
          instancesStore = _this.db.transaction(['instances'], 'readwrite').objectStore('instances');
          data = {};
          return cursor = instancesStore.openCursor().onsuccess = function(e) {
            var changes, instance, storeVersion, _name;
            cursor = e.target.result;
            if (cursor) {
              instance = cursor.value;
              delete instance._key;
              if (data[_name = instance._model] == null) {
                data[_name] = {};
              }
              data[instance._model][instance._id] = instance;
              delete instance._model;
              delete instance._id;
              return cursor["continue"]();
            } else {
              storeVersion = localStorage.getItem('version');
              if (storeVersion !== version) {
                if (version === '2') {
                  if (storeVersion === 'v1') {
                    changes = localStorage.getItem("" + db + ".changes");
                    if (changes) {
                      changes = JSON.parse(changes);
                      localStorage.setItem("" + db + ".changes", JSON.stringify({
                        1: changes
                      }));
                    }
                  }
                }
                localStorage.setItem('version', version);
              }
              return cb({
                data: data,
                clientId: localStorage.getItem('clientId'),
                schema: JSON.parse(localStorage.getItem("" + db + ".schema")),
                changes: JSON.parse(localStorage.getItem("" + db + ".changes")),
                pendingUpdate: JSON.parse(localStorage.getItem('pendingUpdate'))
              });
            }
          };
        };
      })(this));
    };

    return OfflineDb;

  })();
});

//# sourceMappingURL=OfflineDB.js.map
