define ['xyzp/UILayer', 'app/UIFrame', 'app/Component'], (UILayer, UIFrame, Component) ->
	UIHelpers = 
		form:
			setErrors: (formEl, errors) ->
						formEl.find('.errorText').remove()
						formEl.find("[name], .field, .fieldGroup").removeClass 'error'
						for field, error of errors
							fieldEl = formEl.find("[name=#{field}], .field.#{field}, .fieldGroup.#{field}")
							fieldEl.addClass 'error'
							fieldEl.after "<span class='errorText'>#{error}</span>"

			validate: (formEl, val, cb) ->
				errors = {}
				waiting = 0

				done = ->
					if !waiting || !--waiting
						UIHelpers.form.setErrors formEl, errors
						cb _.isEmpty errors

				if _.isFunction val
					val formEl, errors
				else
					for field, v of val
						errorField = field
						if !_.isFunction v
							if _.isPlainObject v
								errorField = v.field
								v = v.validate

							v = switch v
								when 'notEmpty' then (value) -> if !value then 'Please enter value.'

						e = v formEl.find("[name=#{field}]").val(), formEl.find("[name=#{field}]")

						if _.isString e
							errors[errorField] = e

				if !waiting
					done()

		inputData: (el, mapping={}) ->
			data = {}
			el.find('[name]').each ->
				inputEl = $ @
				name = inputEl.attr 'name'
				if inputEl.is('input') && inputEl.attr('type') in ['text', 'email', 'password'] || inputEl.is('select') || inputEl.is('textarea')
					data[name] = if mapping[name] then mapping[name] inputEl.val() else inputEl.val()
				else if inputEl.is('input') && inputEl.attr('type') == 'checkbox'
					data[name] = inputEl.prop 'checked'
			data

		setInputData: (el, data, mapping={}) ->
			for name, value of data
				inputEl = el.find("[name=#{name}]")
				if inputEl.is('input') && inputEl.attr('type') in ['text', 'email', 'password'] || inputEl.is('select') || inputEl.is('textarea')
					inputEl.val if mapping[name] then mapping[name] value else value
				else if inputEl.is('input') && inputEl.attr('type') == 'checkbox'
					inputEl.prop 'checked', !!value
		showDialog: (opts) ->
			dialogEl = $ opts.html
			close = ->
				backgroundEl.fadeOut -> backgroundEl.remove(); clearInterval id

			updatePosition = ->
				dialogEl.css
					left:$('.dialogWrapper').width()/2 - dialogEl.outerWidth()/2
					top:Math.max 0, $('.dialogWrapper').height()/2 - dialogEl.outerHeight()/2

			opts.init el:dialogEl, close:close, updatePosition:updatePosition
			backgroundEl = $('<div class="dialogWrapper" />').appendTo 'body'
			dialogEl.appendTo backgroundEl


			updatePosition()
			id = setInterval updatePosition, 100

		showModalBox: (clientEl, opts={}) ->
			alreadyOpen = @modalBoxData?

			if !@modalBoxData
				console.log rootFrame.chain, rootFrame.chain.length
				@modalBoxData =
					linkNumber:rootFrame.chain.length
					layer:rootLayer.addLayer new UILayer
					close: (cb) =>
						rootFrame.unlinkAt @modalBoxData.linkNumber
						@modalBoxData.backgroundEl.fadeOut =>
							@modalBoxData.layer.destruct()
					backgroundEl: $('<div class="dialogWrapper" />')
					updatePosition: =>
						if @modalBoxData.modalBox.opts.autoPositioned != false
							@modalBoxData.modalBox.el.css
								left:$('.dialogWrapper').width()/2 - @modalBoxData.modalBox.el.outerWidth()/2
								top:Math.max 0, $('.dialogWrapper').height()/2 - @modalBoxData.modalBox.el.outerHeight()/2

				@modalBoxData.layer.addDestructFunc =>
					clearInterval id
					delete @modalBoxData

				@modalBoxData.layer.addHtml $('body'), @modalBoxData.backgroundEl

				id = setInterval (=>@modalBoxData.updatePosition()), 100

			prevModalBox = @modalBoxData.modalBox
			@modalBoxData.modalBox = opts:opts, el:clientEl, close:(=>@modalBoxData.close()), layer:@modalBoxData.layer.addLayer new UILayer

			if alreadyOpen
				prevModalBox.el.fadeOut =>
					prevModalBox.layer.destruct()
					@modalBoxData.modalBox.layer.addHtml @modalBoxData.backgroundEl, clientEl
					@modalBoxData.updatePosition()

				@modalBoxData.backgroundEl.removeClass 'staticClient'
			else
				@modalBoxData.modalBox.layer.addHtml @modalBoxData.backgroundEl, clientEl
				@modalBoxData.updatePosition()

			if opts.absolutePositioned == false
				@modalBoxData.backgroundEl.addClass 'staticClient'

			@modalBoxData.modalBox
		

		editDialog: (opts) ->
			->
				instance = opts.instance
				UIHelpers.showDialog
					html:opts.html()
					init: (el:el, close:close) ->
						if instance
							data = {}
							for field in opts.fields
								data[field] = instance.get field

							UIHelpers.setInputData el, data, opts.display

							el.find('.delete').click ->
								instance.delete()
								close()

						if opts.mask
							for field, mask of opts.mask
								el.find("[name=#{field}]").inputmask mask

						el.find('input[type=submit]').click ->
							data = UIHelpers.inputData el, opts.filter
							if instance
								instance.setProperties data
							else
								dataManager.model(opts.model).create _.merge data, user:g_userCont.get()
							close()
							false
						el.find('.cancel').click close

						opts.init? el

		editDialogBinding: (html, model, fields, init=null) ->
			(instance) ->
				UIHelpers.editDialog(html:html, model:model, fields:fields, instance:instance, init:((el) -> init? el, instance))()

		openModelView: (plate) ->
			el = $('<div class="modalView">
				<div class="header">
					<a href="#" class="back" />
					<div class="title" />
					<a href="#" class="close" />
				</div>
				<div class="client" />
			</div>')

			el.find('.header .title')
				.html plate.title
				.css 'backgroundImage', "url('app/images/#{plate.icon}')"

			el.find('.client').html "<div class='#{plate.class}'>#{plate.html}</div>"

			# close = ->
			# 	el.fadeOut ->
			# 		el.remove()
			# 		layer.destruct()

			el.find('.close').click -> modalBox.close(); false

			layer = new UILayer


			component = new Component layer, el.find('.client').children(':first'), plate.componentIniter
			component.init()
			component.listen (signal) ->
				if signal == 'close'
					close()

			component.listen plate.listener

			modalBox = UIHelpers.showModalBox el, autoPositioned:false


			frame = new UIFrame
			plate.initFrame? frame


			modalBox.layer.addLayer layer
			frame.uiInited()
